#!/usr/bin/perl
# Eelco Mossel, May-August 2006

## Takes a text that has been tokenized and annotated by tokenize.pl and text2lat.pl from CDG and the XML output of the german CDG grammar.
## This script needs as first argument: file.ann (the stand-off annotation) and as second argument: file.cdg (the tokenized file in sentence form)

## The stand-off annotation file looks like this:
#<?xml version="1.0" encoding="ISO-8859-1"?>
#<ROOT>
#    <annotation id="s1" lattice="s1" nowords="6" noedges="12">
#      <file name="(null)" lineNo="0"/>
#      <arc from="0" to="1" word="warum">
#        <tag cat="PWAV"/>
#        <dep level="SYN" label="ADV" modifiee="2"/>
#        <dep level="REF" label="" modifiee="0"/>
#      </arc>
#      <arc from="1" to="2" word="mssen">
#        <tag base="mssen"/>
#        <tag cat="VMFIN"/>
#        <tag mood="indicative"/>
#        <tag number="pl"/>
#        <tag person="third"/>
#        <tag tense="present"/>
#        <dep level="SYN" label="S" modifiee="0"/>
#        <dep level="REF" label="" modifiee="0"/>
#      </arc>

## The tokenized file in sentence form looks like this:
#'s1' : Warum, mssen, Daten, gespeichert, werden, '?';
#'s2' : Durch, den, Speichervorgang, werden, die, von, Ihnen, erzeugten, Daten, '-', 'z.', 'B.', Texte, ',', Tabellen, oder, Bilder, '-', vom, flchtigen, Arbeitsspeicher, in, eine, dauerhaft, existente, Datei, bertragen, '.';

# The tokens from the tokenized file are used to generate the inline-annotated text. The value of the <arc word=""> feature of the analyzed annotation is not used, because this represents the word that is actually found in the lexicon; this is usually a lowercase word, if the surface form is coincidentally (start of sentence) uppercase; in a compound word, it may be just the head. It is thus not necessarily equal to the surface form

## The result must look like this:
#<?xml version="1.0" encoding="ISO-8859-1"?>
#<!DOCTYPE LT4ELAna>
#<LT4ELAna version="1.0project">
#<s id="s1">
#<tok id="t1" class="word" base="warum" ctag="PWAV" msd="0,0,0,0,0,0,0,0">Warum</tok>
#<tok id="t2" class="word" base="mssen" ctag="VMFIN" msd="pl,0,0,third,0,present,0,0">mssen</tok><chunk id="c1" category="NP">
#<tok id="t3" class="word" base="Datum" ctag="NN" msd="pl,neut,bot,third,0,0,0,0">Daten</tok></chunk>
#<tok id="t4" class="word" base="speichern" ctag="VVPP" msd="0,0,0,0,0,0,0,0">gespeichert</tok>
#<tok id="t5" class="word" base="werden" ctag="VAINF" msd="0,0,0,0,0,0,0,0">werden</tok>
#<tok id="t6" class="punct" base="?" ctag="$." msd="0,0,0,0,0,0,0,0">?</tok></s>
#<s id="s2"><tok id="t7" ...........
#..........
#</LT4ELAna>


use strict;
use warnings;

use XML::Parser

open(CDG, $ARGV[1]);
my @sentences = <CDG>; #tokenized file in sentence form

$parser::XML::Parser = new XML::Parser(Handlers => { Start => \&handle_start_firsttime });
$parser::XML::Parser->parsefile($ARGV[0]);

# global
my %cats; # associative array containing sen_no-tok_no => POS-category; e.g.: 1-2 => "VMFIN"
my %modifiees; # associative array containing sen_no-tok_no => sen_no-tok_no of the token which is "modified" (dependency structure) by this one
my $np_status = 3; # used for marking NPs. 0 == begin, 1 == within, 2 == end, 3 == no NP, 4 == one-word NP
my $tok_output; #string. used for accumulating what is written to the output for each token
my $word_id; #integer. number of the word/token in the sentence
my $tok_class; #string. assigned "word" or "punct" and used as value for <tok> element.
my $lexicon_word; #string. contains the value of the word-feature of the currently/last processed token

## counters for IDs for chunks and tokens. Everytime they are used, 1 is added, so that every time a unique one is assigned to a tag
my $chunk_id = 1;
my $tok_id = 1;

binmode(STDOUT, ':encoding(ISO-8859-1)');

#<?xml version="1.0" encoding="UTF-8"?>
#<!DOCTYPE LT4ELAna>
#<LT4ELAna version="1.0project">
print "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>"
    . "\n<!DOCTYPE LT4ELAna>"
    . "\n<LT4ELAna version=\"1.0project\">\n" ;
$parser::XML::Parser = new XML::Parser(Handlers => {Start => \&handle_start,
                                                    End   => \&handle_end    });
$parser::XML::Parser->parsefile($ARGV[0]);
print "\n</LT4ELAna>\n";


# scope could be local but value must be preserved over multiple function calls
my $sentence_id; 
# parse the xml file a first time for storing values to the arrays %cats and %modifiees, which are used as a "look-ahead" in the second run
sub handle_start_firsttime {
  my $element_name; my @avl; my $cat;
  ($_, $element_name, @avl) = @_; #avl = attribute-value-list
  if ($element_name eq "annotation") {
    # <annotation id="s1" lattice="s1" nowords="6" noedges="12">
    $sentence_id = $avl[1];
  }
  elsif ($element_name eq "arc") {
    # <arc from="1" to="2" word="mssen">
    $word_id = $avl[3];
  }
  elsif ($element_name eq "tag" && $avl[0] eq "cat") {
    # <tag cat="VMFIN"/>
    $cat = $avl[1];
    $cats{$sentence_id."-".$word_id} = $cat;
  }
  elsif ($element_name eq "dep" && $avl[0] eq "level" && $avl[1] eq "SYN") {
    $modifiees{$sentence_id."-".$word_id} = $sentence_id."-".$avl[5];
  }
}

## Recursive method
## Returns the identifier (e.g. s2-5) of the end of an NP, if the specified argument is the beginning of an NP
## Looks whether this token modifies (in the syntactic dependency structure), possibly indirectly an NN occurring later in the sentence.
## This means, no PPs following the head noun are included.
sub find_NP { #specify argument: sentence_id-word_id
  my $id; my $modi; my $word_no; my $modi_word_no;
  $id = $_[0];
  ($_, $word_no) = split(/-/, $id);
  $modi = $modifiees{$id};
  ($_, $modi_word_no) = split(/-/, $modi);
  if ($cats{$id} eq "NN")          { return $id; }
  elsif ($modi_word_no < $word_no) { return "s-0";}
  elsif ($cats{$modi} eq "NN")     { return $modi; }
  else                             { return find_NP($modi); }
}

# scope could be local but value must be preserved over multiple function calls
my $found_np;   #integer. used to hold the token-number of the last token of an NP, returned by find_NP
my $current_sentence; #string
my @current_tokens; # used to hold the tokens of the sentence currently processed
my $base_form;
my $ctag;
my $number;
my $gender;
my $case;
my $person;
my $degree;
my $tense;
my $stress;
my $subcat;
sub handle_start {
  my $element_name; my @avl;
  ($_, $element_name, @avl) = @_;
  if ($element_name eq "annotation") {
    # <annotation id="s1" lattice="s1" nowords="6" noedges="12">
    $sentence_id = $avl[1];
    $current_sentence = $sentences[substr($sentence_id, 1) -1]; #first line (index 0) of tokenized file is s1
    $current_sentence =~ s/'s\d+' : //; ## remove the sentence label, e.g. 's1' : Warum, mssen, --> Warum, mssen,
    $current_sentence = substr($current_sentence, 0, -2); # remove the semicolon from the end
    # Split into tokens by delimiter ", " (comma followed by space), but regard commas in quoted tokens such as 'a, b'
    # Temporarily replace those by a one-byte unicode character that is not in the latin1 set (choose ones from cp-1252)
    # Use an opening and a closing placeholder, to prevent recognizing a comma in between in e.g.: 'first-special', '?'
    # The quote must be either the beginning of a line, or must be preceded by a character that is not a backslash (it must be
    ## non-escaped. However, an escaped backslash can occur in front of it, so do allow an even number of backslashes. 
    while ($current_sentence =~ s/(^|[^\\]|(\\\\)+?)'/$1\x{2039}/) {
           $current_sentence =~ s/([^\\]|(\\\\)+?)'/$1\x{203a}/; # I assume an even number of non-escaped quotes
    }
    # if a comma followed by a space occurs within a quoted context, replace the space by a placeholder
    # as lazy as possible using *? in order not to match too large contexts 
    while ($current_sentence =~ s/(\x{2039}[^\x{2039}\x{203a}]*?), ([^\x{2039}\x{203a}]*?\x{203a})/$1,\x{2022}$2/g) {}
    @current_tokens = split(/\, /, $current_sentence); # split into tokens by delimiter ", " (comma followed by space)
    foreach my $t (@current_tokens) {
      $t =~ s/^\x{2039}(.*?)\x{203a}$/$1/; # remove single quotes that are around some tokens, e.g. ',' --> ,
      $t =~ s/\\('|"|\\)/$1/g; # decode escaped ' and " and \
      $t =~ tr/\x{2022}/ /; #change back spaces
    }
    print "<s id=\"".$sentence_id."\">"; # e.g. <s id="s2">
  }
  elsif ($element_name eq "arc") {
    #set variables to "0"; will be filled/changed when processing the corresponding embedded tag elements, if present
    $number = "0";
    $gender = "0";
    $case   = "0";
    $person = "0";
    $degree = "0";
    $tense  = "0";
    $stress = "0";
    $subcat = "0";
    # <arc from="1" to="2" word="mssen">
    $word_id = $avl[3];
    $tok_output = encode_XML_entities($current_tokens[$word_id -1]); 
    $base_form = "";
    $lexicon_word = $avl[5]; ## used as base form if no <tag base> is present
  }
  elsif ($element_name eq "tag") {
    ##  <tag base="mssen"/>
    ##  <tag cat="VMFIN"/>
    ##  <tag mood="indicative"/>
    ##  <tag number="pl"/>
    ##  <tag person="third"/>
    ##  <tag tense="present"/>
    if ($avl[0] eq "base") {
      $base_form = encode_XML_entities($avl[1]);
    }
    elsif ($avl[0] eq "cat") {
      if ($base_form eq "") { #this only works because the tags are ordered alphabetically, so if base is present, 
                                   # it has been found before cat
        $base_form = encode_XML_entities($lexicon_word);
      }
      $ctag = $avl[1]; #ctag is always present
      if ($ctag =~ /^\$/) { # part of speech == $, or $. or $(
        $tok_class = "punct";
      }
      else {
        $tok_class = "word";
      }
    }
    else { # the rest are morphosyntactic features. collect the used ones in the corresponding variables
      if    ($avl[0] eq "number") { $number = $avl[1]; }
      elsif ($avl[0] eq "gender") { $gender = $avl[1]; }
      elsif ($avl[0] eq "case")   { $case   = $avl[1]; }
      elsif ($avl[0] eq "person") { $person = $avl[1]; }
      elsif ($avl[0] eq "degree") { $degree = $avl[1]; }
      elsif ($avl[0] eq "tense")  { $tense  = $avl[1]; }
      elsif ($avl[0] eq "stress") { $stress = $avl[1]; }
      elsif ($avl[0] eq "subcat") { $subcat = $avl[1]; }
    }
  }
  elsif ($element_name eq "dep" && $avl[0] eq "level" && $avl[1] eq "SYN") {
    # <dep level="SYN" label="DET" modifiee="3"/>
    if ($np_status == 3) {
      ($_, $found_np) = split(/-/, find_NP($sentence_id."-".$word_id));
      if ($found_np > 0) { $np_status = 0; } # 0 == begin, 1 == within, 2 == end, 3 == no NP, 4 == one-word NP
    }
  }
  if ($np_status == 1 && $found_np == $word_id) {
    $np_status = 2;
  }
  elsif ($np_status == 0 && $found_np == $word_id) { #one-word NP
    $np_status = 4;
  }
}

sub encode_XML_entities {
  my $arg = $_[0];
  $arg =~ s/&/&amp\;/g; # caution: this must be done first, because the next lines generate more ampersands
  $arg =~ s/</&lt\;/g;
  $arg =~ s/>/&gt\;/g;
  $arg =~ s/\"/&quot\;/g;
  $arg =~ s/\'/&apos\;/g;
  return $arg;
}

sub handle_end {
  my $element_name;
  ($_, $element_name) = @_;
  if ($element_name eq "annotation") {
    print "</s>";
  }
  if ($element_name eq "arc") {
    ##<tok id="t1" class="word" base="Kurzanleitung" ctag="NN" msd="sg,fem,bot,third,0,0,0,0">Kurzanleitung</tok></chunk>
    my $tok_tag = "<tok id=\"t".$tok_id++."\" class=\"".$tok_class.
        #"\" base=\"".$base_form."\" ctag=\"".$ctag.
        #"\" msd=\"".$number.",".$gender.",".$case.",".$person.",".$degree.",".$tense.",".$stress.",".$subcat."\">";
        # adapted for education course
        "\" base=\"".$base_form."\" ctag=\"".$ctag."\">";
    $tok_output = "\n".$tok_tag.$tok_output."</tok>";
    if ($np_status == 0) {
      my $np_tag = "<chunk id=\"c".$chunk_id++."\" category=\"NP\">";
      print $np_tag . $tok_output;
      $np_status = 1;
    }
    elsif ($np_status == 1 || $np_status == 3) {
      print $tok_output;
    }
    elsif ($np_status == 2) {
      print $tok_output . "</chunk>";
      $np_status = 3;
    }
    elsif ($np_status == 4) {
      my $np_tag = "<chunk id=\"c".$chunk_id++."\" category=\"NP\">";
      print $np_tag . $tok_output . "</chunk>";
      $np_status = 3;
    }
  } # end if "arc"
}
